"""
at the moment the endpoint command lets you specify the api and s3 endpoints
eventually it will be sufficient to just specify the api endpoint and the s3 endpoint will
be provided by the api
"""

from __future__ import annotations

from typing import Optional

import typer
from rich.console import Console

from kleinkram.config import Endpoint
from kleinkram.config import add_endpoint
from kleinkram.config import endpoint_table
from kleinkram.config import get_config
from kleinkram.config import select_endpoint

HELP = """\
Get or set the current endpoint.

The endpoint is used to determine the API server to connect to\
(default is the API server of https://datasets.leggedrobotics.com).
"""

endpoint_typer = typer.Typer(
    name="endpoint",
    help=HELP,
    context_settings={"help_option_names": ["-h", "--help"]},
    invoke_without_command=True,
)


@endpoint_typer.callback()
def endpoint(
    name: Optional[str] = typer.Argument(None, help="Name of the endpoint to use"),
    api: Optional[str] = typer.Argument(None, help="API endpoint to use"),
    s3: Optional[str] = typer.Argument(None, help="S3 endpoint to use"),
) -> None:
    config = get_config()
    console = Console()

    if not any([name, api, s3]):
        console.print(endpoint_table(config))
    elif name is not None and not any([api, s3]):
        try:
            select_endpoint(config, name)
        except ValueError:
            console.print(f"Endpoint {name} not found.\n", style="red")
            console.print(endpoint_table(config))
    elif not (name and api and s3):
        raise typer.BadParameter(
            "to add a new endpoint you must specify the api and s3 endpoints"
        )
    else:
        new_endpoint = Endpoint(name, api, s3)
        add_endpoint(config, new_endpoint)



from __future__ import annotations

from typing import Optional

import typer

import kleinkram.api.routes
import kleinkram.core
from kleinkram.api.client import AuthenticatedClient
from kleinkram.api.query import FileQuery
from kleinkram.api.query import MissionQuery
from kleinkram.api.query import ProjectQuery
from kleinkram.api.routes import get_file
from kleinkram.config import get_shared_state
from kleinkram.printing import print_file_info
from kleinkram.utils import split_args

INFO_HELP = "get information about a file"
DELETE_HELP = "delete a file"


file_typer = typer.Typer(
    no_args_is_help=True, context_settings={"help_option_names": ["-h", "--help"]}
)


@file_typer.command(help=INFO_HELP)
def info(
    project: Optional[str] = typer.Option(
        None, "--project", "-p", help="project id or name"
    ),
    mission: Optional[str] = typer.Option(
        None, "--mission", "-m", help="mission id or name"
    ),
    file: str = typer.Option(..., "--file", "-f", help="file id or name"),
) -> None:
    project_ids, project_patterns = split_args([project] if project else [])
    project_query = ProjectQuery(ids=project_ids, patterns=project_patterns)

    mission_ids, mission_patterns = split_args([mission] if mission else [])
    mission_query = MissionQuery(
        ids=mission_ids,
        patterns=mission_patterns,
        project_query=project_query,
    )

    file_ids, file_patterns = split_args([file])
    file_query = FileQuery(
        ids=file_ids,
        patterns=file_patterns,
        mission_query=mission_query,
    )

    client = AuthenticatedClient()
    file_parsed = get_file(client, file_query)
    print_file_info(file_parsed, pprint=get_shared_state().verbose)


@file_typer.command(help=DELETE_HELP)
def delete(
    project: Optional[str] = typer.Option(
        None, "--project", "-p", help="project id or name"
    ),
    mission: Optional[str] = typer.Option(
        None, "--mission", "-m", help="mission id or name"
    ),
    file: str = typer.Option(..., "--file", "-f", help="file id or name"),
    confirm: bool = typer.Option(
        False, "--confirm", "-y", "--yes", help="confirm deletion"
    ),
) -> None:
    if not confirm:
        typer.confirm(f"delete {project} {mission}", abort=True)

    project_ids, project_patterns = split_args([project] if project else [])
    project_query = ProjectQuery(ids=project_ids, patterns=project_patterns)

    mission_ids, mission_patterns = split_args([mission] if mission else [])
    mission_query = MissionQuery(
        ids=mission_ids,
        patterns=mission_patterns,
        project_query=project_query,
    )

    file_ids, file_patterns = split_args([file])
    file_query = FileQuery(
        ids=file_ids,
        patterns=file_patterns,
        mission_query=mission_query,
    )

    client = AuthenticatedClient()
    file_parsed = get_file(client, file_query)
    kleinkram.core.delete_files(client=client, file_ids=[file_parsed.id])



"""
this file contains the main functionality of kleinkram cli

- download
- upload
- verify
- update_file
- update_mission
- update_project
- delete_files
- delete_mission
- delete_project
"""

from __future__ import annotations

from pathlib import Path
from typing import Collection
from typing import Dict
from typing import List
from typing import Optional
from typing import Sequence
from uuid import UUID

from rich.console import Console
from tqdm import tqdm

import kleinkram.api.file_transfer
import kleinkram.api.routes
import kleinkram.errors
from kleinkram.api.client import AuthenticatedClient
from kleinkram.api.query import FileQuery
from kleinkram.api.query import MissionQuery
from kleinkram.api.query import ProjectQuery
from kleinkram.api.query import check_mission_query_is_creatable
from kleinkram.errors import MissionNotFound
from kleinkram.models import FileState
from kleinkram.models import FileVerificationStatus
from kleinkram.printing import files_to_table
from kleinkram.utils import b64_md5
from kleinkram.utils import check_file_paths
from kleinkram.utils import file_paths_from_files
from kleinkram.utils import get_filename_map


def download(
    *,
    client: AuthenticatedClient,
    query: FileQuery,
    base_dir: Path,
    nested: bool = False,
    overwrite: bool = False,
    verbose: bool = False,
) -> None:
    """\
    downloads files, asserts that the destition dir exists
    returns the files that were downloaded

    TODO: the above is a lie, at the moment we just return all files that were found
    this might include some files that were skipped or not downloaded for some reason
    we would need to modify the `download_files` function to return this in the future
    """

    if not base_dir.exists():
        raise ValueError(f"Destination {base_dir.absolute()} does not exist")
    if not base_dir.is_dir():
        raise ValueError(f"Destination {base_dir.absolute()} is not a directory")

    # retrive files and get the destination paths
    files = list(kleinkram.api.routes.get_files(client, file_query=query))
    paths = file_paths_from_files(files, dest=base_dir, allow_nested=nested)

    if verbose:
        table = files_to_table(files, title="downloading files...")
        Console().print(table)

    kleinkram.api.file_transfer.download_files(
        client, paths, verbose=verbose, overwrite=overwrite
    )


def upload(
    *,
    client: AuthenticatedClient,
    query: MissionQuery,
    file_paths: Sequence[Path],
    create: bool = False,
    metadata: Optional[Dict[str, str]] = None,
    ignore_missing_metadata: bool = False,
    verbose: bool = False,
) -> None:
    """\
    uploads files to a mission

    create a mission if it does not exist if `create` is True
    in that case you can also specify `metadata` and `ignore_missing_metadata`
    """
    # check that file paths are for valid files and have valid suffixes
    check_file_paths(file_paths)

    try:
        mission = kleinkram.api.routes.get_mission(client, query=query)
    except MissionNotFound:
        if not create:
            raise
        mission = None

    if create and mission is None:
        # check if project exists and get its id at the same time
        project_id = kleinkram.api.routes.get_project(
            client, query=query.project_query
        ).id
        mission_name = check_mission_query_is_creatable(query)
        kleinkram.api.routes._create_mission(
            client,
            project_id,
            mission_name,
            metadata=metadata or {},
            ignore_missing_tags=ignore_missing_metadata,
        )
        mission = kleinkram.api.routes.get_mission(client, query)

    assert mission is not None, "unreachable"

    filename_map = get_filename_map(file_paths)
    kleinkram.api.file_transfer.upload_files(
        client, filename_map, mission.id, verbose=verbose
    )


def verify(
    *,
    client: AuthenticatedClient,
    query: MissionQuery,
    file_paths: Sequence[Path],
    skip_hash: Optional[bool] = None,
    check_file_hash: bool = True,
    check_file_size: bool = False,
    verbose: bool = False,
) -> Dict[Path, FileVerificationStatus]:

    # add deprecated warning for skip_hash
    if skip_hash is not None:
        print(
            "Warning: --skip-hash is deprecated and will be removed in a future version. "
            "Use --check-file-hash=False instead.",
        )
        check_file_hash = not skip_hash

    # check that file paths are for valid files and have valid suffixes
    check_file_paths(file_paths)

    # check that the mission exists
    _ = kleinkram.api.routes.get_mission(client, query)

    remote_files = {
        f.name: f
        for f in kleinkram.api.routes.get_files(
            client, file_query=FileQuery(mission_query=query)
        )
    }
    filename_map = get_filename_map(file_paths)

    # verify files
    file_status: Dict[Path, FileVerificationStatus] = {}
    for name, file in tqdm(
        filename_map.items(),
        desc="verifying files",
        unit="file",
        disable=not verbose,
    ):
        if name not in remote_files:
            file_status[file] = FileVerificationStatus.MISSING
            continue

        remote_file = remote_files[name]

        if remote_file.state == FileState.UPLOADING:
            file_status[file] = FileVerificationStatus.UPLOADING
        elif remote_file.state == FileState.OK:

            # default case, will be overwritten if we find a mismatch
            file_status[file] = FileVerificationStatus.UPLOADED

            if check_file_size:
                if remote_file.size == file.stat().st_size:
                    file_status[file] = FileVerificationStatus.UPLOADED
                else:
                    file_status[file] = FileVerificationStatus.MISMATCHED_SIZE

            if file_status[file] != FileVerificationStatus.UPLOADED:
                continue  # abort if we already found a mismatch

            if check_file_hash:
                if remote_file.hash is None:
                    file_status[file] = FileVerificationStatus.COMPUTING_HASH
                elif remote_file.hash == b64_md5(file):
                    file_status[file] = FileVerificationStatus.UPLOADED
                else:
                    file_status[file] = FileVerificationStatus.MISMATCHED_HASH

        else:
            file_status[file] = FileVerificationStatus.UNKNOWN

    return file_status


def update_file(*, client: AuthenticatedClient, file_id: UUID) -> None:
    """\
    TODO: what should this even do
    """
    _ = client, file_id
    raise NotImplementedError("if you have an idea what this should do, open an issue")


def update_mission(
    *, client: AuthenticatedClient, mission_id: UUID, metadata: Dict[str, str]
) -> None:
    # TODO: this funciton will do more than just overwirte the metadata in the future
    kleinkram.api.routes._update_mission(client, mission_id, metadata=metadata)


def update_project(
    *,
    client: AuthenticatedClient,
    project_id: UUID,
    description: Optional[str] = None,
    new_name: Optional[str] = None,
) -> None:
    # TODO: this function should do more in the future
    kleinkram.api.routes._update_project(
        client, project_id, description=description, new_name=new_name
    )


def delete_files(*, client: AuthenticatedClient, file_ids: Collection[UUID]) -> None:
    """\
    deletes multiple files accross multiple missions
    """
    if not file_ids:
        return

    # we need to check that file_ids is not empty, otherwise this is bad
    files = list(kleinkram.api.routes.get_files(client, FileQuery(ids=list(file_ids))))

    # check if all file_ids were actually found
    found_ids = [f.id for f in files]
    for file_id in file_ids:
        if file_id not in found_ids:
            raise kleinkram.errors.FileNotFound(
                f"file {file_id} not found, did not delete any files"
            )

    # to prevent catastrophic mistakes from happening *again*
    assert set(file_ids) == set([file.id for file in files]), "unreachable"

    # we can only batch delete files within the same mission
    missions_to_files: Dict[UUID, List[UUID]] = {}
    for file in files:
        if file.mission_id not in missions_to_files:
            missions_to_files[file.mission_id] = []
        missions_to_files[file.mission_id].append(file.id)

    for mission_id, ids_ in missions_to_files.items():
        kleinkram.api.routes._delete_files(client, file_ids=ids_, mission_id=mission_id)


def delete_mission(*, client: AuthenticatedClient, mission_id: UUID) -> None:
    mquery = MissionQuery(ids=[mission_id])
    mission = kleinkram.api.routes.get_mission(client, mquery)
    files = list(
        kleinkram.api.routes.get_files(
            client, file_query=FileQuery(mission_query=mquery)
        )
    )

    # delete the files and then the mission
    kleinkram.api.routes._delete_files(client, [f.id for f in files], mission.id)
    kleinkram.api.routes._delete_mission(client, mission_id)


def delete_project(*, client: AuthenticatedClient, project_id: UUID) -> None:
    pquery = ProjectQuery(ids=[project_id])
    _ = kleinkram.api.routes.get_project(client, pquery)  # check if project exists

    # delete all missions and files
    missions = list(
        kleinkram.api.routes.get_missions(
            client, mission_query=MissionQuery(project_query=pquery)
        )
    )
    for mission in missions:
        delete_mission(client=client, mission_id=mission.id)

    # delete the project
    kleinkram.api.routes._delete_project(client, project_id)



from __future__ import annotations

import logging
import os
import sys
import time
from enum import Enum
from pathlib import Path
from typing import List
from typing import Optional

import typer
from click import Context
from rich.console import Console
from typer.core import TyperGroup

from kleinkram._version import __version__
from kleinkram.api.client import AuthenticatedClient
from kleinkram.api.routes import _claim_admin
from kleinkram.api.routes import _get_api_version
from kleinkram.auth import login_flow
from kleinkram.cli._download import download_typer
from kleinkram.cli._endpoint import endpoint_typer
from kleinkram.cli._file import file_typer
from kleinkram.cli._list import list_typer
from kleinkram.cli._mission import mission_typer
from kleinkram.cli._project import project_typer
from kleinkram.cli._upload import upload_typer
from kleinkram.cli._verify import verify_typer
from kleinkram.cli.error_handling import ErrorHandledTyper
from kleinkram.cli.error_handling import display_error
from kleinkram.config import MAX_TABLE_SIZE
from kleinkram.config import Config
from kleinkram.config import check_config_compatibility
from kleinkram.config import get_config
from kleinkram.config import get_shared_state
from kleinkram.config import save_config
from kleinkram.errors import InvalidCLIVersion
from kleinkram.utils import format_traceback
from kleinkram.utils import get_supported_api_version

# slightly cursed lambdas so that linters don't complain about unreachable code
if (lambda: os.name)() == "posix":
    LOG_DIR = Path().home() / ".local" / "state" / "kleinkram"
elif (lambda: os.name)() == "nt":
    LOG_DIR = Path().home() / "AppData" / "Local" / "kleinkram"
else:
    raise OSError(f"Unsupported OS {os.name}")

LOG_FILE = LOG_DIR / f"{time.time_ns()}.log"
LOG_FORMAT = "%(asctime)s | %(name)s | %(levelname)s | %(message)s"

# setup default logging
logger = logging.getLogger(__name__)


CLI_HELP = """\
Kleinkram CLI

The Kleinkram CLI is a command line interface for Kleinkram.
For a list of available commands, run `klein --help` or visit \
https://docs.datasets.leggedrobotics.com/usage/python/getting-started.html \
for more information.
"""


class LogLevel(str, Enum):
    DEBUG = "DEBUG"
    INFO = "INFO"
    WARNING = "WARNING"
    ERROR = "ERROR"
    CRITICAL = "CRITICAL"


LOG_LEVEL_MAP = {
    LogLevel.DEBUG: logging.DEBUG,
    LogLevel.INFO: logging.INFO,
    LogLevel.WARNING: logging.WARNING,
    LogLevel.ERROR: logging.ERROR,
    LogLevel.CRITICAL: logging.CRITICAL,
}


class CommandTypes(str, Enum):
    AUTH = "Authentication Commands"
    CORE = "Core Commands"
    CRUD = "Create Update Delete Commands"


class OrderCommands(TyperGroup):
    def list_commands(self, ctx: Context) -> List[str]:
        _ = ctx  # suppress unused variable warning
        return list(self.commands)


app = ErrorHandledTyper(
    cls=OrderCommands,
    help=CLI_HELP,
    context_settings={"help_option_names": ["-h", "--help"]},
    no_args_is_help=True,
)

app.add_typer(endpoint_typer, name="endpoint", rich_help_panel=CommandTypes.AUTH)

app.add_typer(download_typer, name="download", rich_help_panel=CommandTypes.CORE)
app.add_typer(upload_typer, name="upload", rich_help_panel=CommandTypes.CORE)
app.add_typer(verify_typer, name="verify", rich_help_panel=CommandTypes.CORE)
app.add_typer(list_typer, name="list", rich_help_panel=CommandTypes.CORE)

app.add_typer(file_typer, name="file", rich_help_panel=CommandTypes.CRUD)
app.add_typer(mission_typer, name="mission", rich_help_panel=CommandTypes.CRUD)
app.add_typer(project_typer, name="project", rich_help_panel=CommandTypes.CRUD)


# attach error handler to app
@app.error_handler(Exception)
def base_handler(exc: Exception) -> int:
    shared_state = get_shared_state()

    display_error(exc=exc, verbose=shared_state.verbose)
    logger.error(format_traceback(exc))

    if not shared_state.debug:
        return 1
    raise exc


@app.command(rich_help_panel=CommandTypes.AUTH)
def login(
    oAuthProvider: str = typer.Option(
        "google",
        "--oauth-provider",
        "-p",
        help="OAuth provider to use for login. Supported providers: google, github, fake-oauth.",
        show_default=True,
    ),
    key: Optional[str] = typer.Option(None, help="CLI key"),
    headless: bool = typer.Option(False),
) -> None:

    # validate oAuthProvider
    if oAuthProvider not in ["google", "github", "fake-oauth"]:
        raise typer.BadParameter(
            f"Unsupported OAuth provider '{oAuthProvider}'. Supported providers: google, github, fake-oauth."
        )

    login_flow(oAuthProvider=oAuthProvider, key=key, headless=headless)


@app.command(rich_help_panel=CommandTypes.AUTH)
def logout(all: bool = typer.Option(False, help="logout on all enpoints")) -> None:
    config = get_config()
    if all:
        config.endpoint_credentials.clear()
    else:
        config.endpoint_credentials.pop(config.selected_endpoint, None)
    save_config(config)


@app.command(hidden=True)
def claim():
    client = AuthenticatedClient()
    _claim_admin(client)
    print("admin rights claimed successfully.")


def _version_callback(value: bool) -> None:
    if value:
        typer.echo(__version__)
        raise typer.Exit()


def check_version_compatiblity() -> None:
    cli_version = get_supported_api_version()
    api_version = _get_api_version()
    api_vers_str = ".".join(map(str, api_version))

    if cli_version[0] != api_version[0]:
        raise InvalidCLIVersion(
            f"CLI version {__version__} is not compatible with API version {api_vers_str}"
        )

    if cli_version[1] != api_version[1]:
        msg = f"CLI version {__version__} might not be compatible with API version {api_vers_str}"
        Console(file=sys.stderr).print(msg, style="red")
        logger.warning(msg)


@app.callback()
def cli(
    verbose: bool = typer.Option(True, help="Enable verbose mode."),
    debug: bool = typer.Option(False, help="Enable debug mode."),
    version: Optional[bool] = typer.Option(
        None, "--version", "-v", callback=_version_callback
    ),
    log_level: Optional[LogLevel] = typer.Option(None, help="Set log level."),
    max_lines: int = typer.Option(
        MAX_TABLE_SIZE,
        "--max-lines",
        help="Maximum number of lines when pretty printing tables. -1 for unlimited.",
    ),
):
    if not check_config_compatibility():
        typer.confirm("found incompatible config file, overwrite?", abort=True)
        save_config(Config())

    _ = version  # suppress unused variable warning
    shared_state = get_shared_state()
    shared_state.verbose = verbose
    shared_state.debug = debug

    if max_lines < 0 and max_lines != -1:
        raise typer.BadParameter("`--max-lines` must be -1 or positive")
    shared_state.max_table_size = max_lines

    if shared_state.debug and log_level is None:
        log_level = LogLevel.DEBUG
    if log_level is None:
        log_level = LogLevel.WARNING

    LOG_DIR.mkdir(parents=True, exist_ok=True)
    logging.basicConfig(
        level=LOG_LEVEL_MAP[log_level], filename=LOG_FILE, format=LOG_FORMAT
    )
    logger.info(f"CLI version: {__version__}")

    try:
        check_version_compatiblity()
    except InvalidCLIVersion as e:
        logger.error(format_traceback(e))
        raise
    except Exception:
        err = "failed to check version compatibility"
        Console(file=sys.stderr).print(
            err, style="yellow" if shared_state.verbose else None
        )
        logger.error(err)



from __future__ import annotations

from secrets import token_hex
from uuid import uuid4

import pytest

import kleinkram.api.routes
import kleinkram.core
import kleinkram.errors
from kleinkram import list_files
from kleinkram import list_missions
from kleinkram import list_projects
from kleinkram.api.client import AuthenticatedClient
from kleinkram.api.query import FileQuery
from kleinkram.api.query import MissionQuery
from kleinkram.api.query import ProjectQuery
from kleinkram.errors import MissionNotFound
from kleinkram.models import FileVerificationStatus
from testing.backend_fixtures import DATA_FILES


@pytest.mark.slow
def test_upload_create(project):
    mission_name = token_hex(8)
    mission_query = MissionQuery(
        patterns=[mission_name], project_query=ProjectQuery(ids=[project.id])
    )

    client = AuthenticatedClient()
    kleinkram.core.upload(
        client=client, query=mission_query, file_paths=DATA_FILES, create=True
    )

    mission = list_missions(mission_names=[mission_name])[0]
    assert mission.project_id == project.id
    assert mission.name == mission_name

    files = list_files(mission_ids=[mission.id])
    assert set([file.name for file in files if file.name.endswith(".bag")]) == set(
        [file.name for file in DATA_FILES]
    )


@pytest.mark.slow
def test_upload_no_create(project):
    mission_name = token_hex(8)
    mission_query = MissionQuery(
        patterns=[mission_name], project_query=ProjectQuery(ids=[project.id])
    )

    client = AuthenticatedClient()
    with pytest.raises(MissionNotFound):
        kleinkram.core.upload(
            client=client, query=mission_query, file_paths=DATA_FILES, create=False
        )


@pytest.mark.slow
def test_upload_to_existing_mission(empty_mission):
    mission_query = MissionQuery(ids=[empty_mission.id])

    client = AuthenticatedClient()
    kleinkram.core.upload(client=client, query=mission_query, file_paths=DATA_FILES)

    files = list_files(mission_ids=[empty_mission.id])
    assert set([file.name for file in files if file.name.endswith(".bag")]) == set(
        [file.name for file in DATA_FILES]
    )


@pytest.mark.slow
def test_delete_existing_files(mission):
    client = AuthenticatedClient()
    files = list_files(mission_ids=[mission.id], file_names=["*.bag"])
    kleinkram.core.delete_files(client=client, file_ids=[f.id for f in files])
    assert not list_files(mission_ids=[mission.id], file_names=["*.bag"])


@pytest.mark.slow
def test_delete_working_as_expected_when_passing_empty_list(mission):
    client = AuthenticatedClient()

    # we need to filter by *.bag to not get flakyness due to conversion
    n_files = len(list_files(mission_ids=[mission.id], file_names=["*.bag"]))
    kleinkram.core.delete_files(client=client, file_ids=[])
    n_files_after_delete = len(
        list_files(mission_ids=[mission.id], file_names=["*.bag"])
    )
    assert n_files == n_files_after_delete


@pytest.mark.slow
def test_delete_non_existing_files():
    client = AuthenticatedClient()

    with pytest.raises(kleinkram.errors.FileNotFound):
        kleinkram.core.delete_files(client=client, file_ids=[uuid4()])


@pytest.mark.slow
def test_create_update_delete_mission(project):
    mission_name = token_hex(8)

    client = AuthenticatedClient()
    kleinkram.api.routes._create_mission(client, project.id, mission_name)

    mission = list_missions(mission_names=[mission_name])[0]

    assert mission.project_id == project.id
    assert mission.name == mission_name

    assert list_files(mission_ids=[mission.id]) == []

    # TODO test update, for this we would need to add metadata types to the backend

    kleinkram.core.delete_mission(client=client, mission_id=mission.id)


@pytest.mark.slow
def test_create_update_delete_project():
    project_name = token_hex(8)

    client = AuthenticatedClient()
    project_id = kleinkram.api.routes._create_project(client, project_name, "test")
    project = list_projects(project_ids=[project_id])[0]

    assert list_missions(project_ids=[project.id]) == []
    assert list_files(project_ids=[project.id]) == []

    assert project.name == project_name
    assert project.description == "test"

    new_name = token_hex(8)
    kleinkram.core.update_project(
        client=client, project_id=project.id, new_name=new_name, description="new desc"
    )

    project = list_projects(project_ids=[project.id])[0]
    assert project.name == new_name
    assert project.description == "new desc"

    kleinkram.core.delete_project(client=client, project_id=project.id)


@pytest.mark.slow
def test_download(mission, tmp_path):
    client = AuthenticatedClient()

    query = FileQuery(mission_query=MissionQuery(ids=[mission.id]), patterns=["*.bag"])
    kleinkram.core.download(client=client, query=query, base_dir=tmp_path)
    files = list_files(mission_ids=[mission.id], file_names=["*.bag"])

    assert set([f.name for f in tmp_path.iterdir()]) == set([f.name for f in files])

    for file in files:
        assert (tmp_path / file.name).stat().st_size == file.size


@pytest.mark.slow
def test_verify(mission):
    client = AuthenticatedClient()
    query = MissionQuery(ids=[mission.id])

    verify_status = kleinkram.core.verify(
        client=client, query=query, file_paths=DATA_FILES, skip_hash=True
    )

    assert all(
        status == FileVerificationStatus.UPLOADED for status in verify_status.values()
    )


@pytest.mark.slow
def test_update_file():
    client = AuthenticatedClient()
    with pytest.raises(NotImplementedError):
        kleinkram.core.update_file(client=client, file_id=uuid4())



from __future__ import annotations

from kleinkram._version import __version__
from kleinkram.wrappers import create_mission
from kleinkram.wrappers import create_project
from kleinkram.wrappers import delete_file
from kleinkram.wrappers import delete_files
from kleinkram.wrappers import delete_mission
from kleinkram.wrappers import delete_project
from kleinkram.wrappers import download
from kleinkram.wrappers import list_files
from kleinkram.wrappers import list_missions
from kleinkram.wrappers import list_projects
from kleinkram.wrappers import update_file
from kleinkram.wrappers import update_mission
from kleinkram.wrappers import update_project
from kleinkram.wrappers import upload
from kleinkram.wrappers import verify

__all__ = [
    "__version__",
    "upload",
    "verify",
    "download",
    "list_files",
    "list_missions",
    "list_projects",
    "update_file",
    "update_mission",
    "update_project",
    "delete_files",
    "delete_file",
    "delete_mission",
    "delete_project",
    "create_mission",
    "create_project",
]



from __future__ import annotations

import json
import sys
from dataclasses import asdict
from datetime import datetime
from pathlib import Path
from typing import List
from typing import Mapping
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import Union

import dateutil.parser
from rich.console import Console
from rich.table import Table
from rich.text import Text

from kleinkram.config import get_shared_state
from kleinkram.core import FileVerificationStatus
from kleinkram.models import File
from kleinkram.models import FileState
from kleinkram.models import MetadataValue
from kleinkram.models import MetadataValueType
from kleinkram.models import Mission
from kleinkram.models import Project

FILE_STATE_COLOR = {
    FileState.OK: "green",
    FileState.CORRUPTED: "red",
    FileState.UPLOADING: "yellow",
    FileState.ERROR: "red",
    FileState.CONVERSION_ERROR: "red",
    FileState.LOST: "bold red",
    FileState.FOUND: "yellow",
}


FILE_VERIFICATION_STATUS_STYLES = {
    FileVerificationStatus.UPLOADED: "green",
    FileVerificationStatus.UPLOADING: "yellow",
    FileVerificationStatus.MISSING: "yellow",
    FileVerificationStatus.MISMATCHED_HASH: "red",
    FileVerificationStatus.MISMATCHED_SIZE: "red",
    FileVerificationStatus.UNKNOWN: "gray",
    FileVerificationStatus.COMPUTING_HASH: "purple",
}


def _add_placeholder_row(table: Table, skipped: int) -> None:
    first_column = f"... ({skipped} more)"
    table.add_row(first_column, *["..." for _ in range(len(table.columns) - 1)])


def file_state_to_text(file_state: FileState) -> Text:
    return Text(file_state.value, style=FILE_STATE_COLOR[file_state])


def file_verification_status_to_text(
    file_verification_status: FileVerificationStatus,
) -> Text:
    return Text(
        file_verification_status.value,
        style=FILE_VERIFICATION_STATUS_STYLES[file_verification_status],
    )


def format_bytes(size: int) -> str:
    """
    Converts a size in bytes to a human-readable string (e.g., KB, MB, GB, TB).

    Args:
        size (int): The size in bytes.

    Returns:
        str: A formatted string representing the size in appropriate units.
    """

    if size < 0:
        raise ValueError("Size must be a non-negative integer")

    # Define units and their thresholds
    units = ["B", "KB", "MB", "GB", "TB", "PB", "EB"]
    index = 0

    fsize: float = size
    while fsize >= 1000 and index < len(units) - 1:
        fsize /= 1000.0
        index += 1

    # Format to 2 decimal places if needed
    if index == 0:  # Bytes don't need decimals
        return f"{int(fsize)} {units[index]}"
    else:
        return f"{fsize:.2f} {units[index]}"


MetadataValueInternalType = Union[str, float, bool, datetime]


def parse_metadata_value(value: MetadataValue) -> Union[str, float, bool, datetime]:
    if value.type_ in [
        MetadataValueType.STRING,
        MetadataValueType.LINK,
        MetadataValueType.LOCATION,
    ]:
        return value.value
    if value.type_ == MetadataValueType.NUMBER:
        return float(value.value)
    if value.type_ == MetadataValueType.BOOLEAN:
        return value.value == "true"
    if value.type_ == MetadataValueType.DATE:
        return dateutil.parser.isoparse(value.value)
    assert False, "unreachable"


def projects_to_table(projects: Sequence[Project]) -> Table:
    table = Table(title="projects")
    table.add_column("id")
    table.add_column("name")
    table.add_column("description")

    max_table_size = get_shared_state().max_table_size
    for project in projects[:max_table_size]:
        table.add_row(str(project.id), project.name, project.description)
    if len(projects) > max_table_size:
        _add_placeholder_row(table, skipped=len(projects) - max_table_size)
    return table


def missions_to_table(missions: Sequence[Mission]) -> Table:
    table = Table(title="missions")
    table.add_column("project")
    table.add_column("name")
    table.add_column("id")
    table.add_column("files")
    table.add_column("size")

    # order by project, name
    missions_tp: List[Tuple[str, str, Mission]] = []
    for mission in missions:
        missions_tp.append((mission.project_name, mission.name, mission))
    missions_tp.sort()

    if not missions_tp:
        return table
    last_project: Optional[str] = None
    max_table_size = get_shared_state().max_table_size
    for project, _, mission in missions_tp[:max_table_size]:
        # add delimiter row if project changes
        if last_project is not None and last_project != project:
            table.add_section()
        last_project = project

        table.add_row(
            mission.project_name,
            mission.name,
            str(mission.id),
            str(mission.number_of_files),
            format_bytes(mission.size),
        )

    if len(missions_tp) > max_table_size:
        _add_placeholder_row(table, skipped=len(missions_tp) - max_table_size)
    return table


def files_to_table(
    files: Sequence[File], *, title: str = "files", delimiters: bool = True
) -> Table:
    table = Table(title=title)
    table.add_column("project")
    table.add_column("mission")
    table.add_column("name")
    table.add_column("id")
    table.add_column("state")
    table.add_column("size")
    table.add_column("categories")

    # order by project, mission, name
    files_tp: List[Tuple[str, str, str, File]] = []
    for file in files:
        files_tp.append((file.project_name, file.mission_name, file.name, file))
    files_tp.sort()

    if not files_tp:
        return table

    last_mission: Optional[str] = None
    max_table_size = get_shared_state().max_table_size
    for _, mission, _, file in files_tp[:max_table_size]:
        if last_mission is not None and last_mission != mission and delimiters:
            table.add_section()
        last_mission = mission

        table.add_row(
            file.project_name,
            file.mission_name,
            file.name,
            Text(str(file.id), style="green"),
            file_state_to_text(file.state),
            format_bytes(file.size),
            ", ".join(file.categories),
        )

    if len(files_tp) > max_table_size:
        _add_placeholder_row(table, skipped=len(files_tp) - max_table_size)

    return table


def file_info_table(file: File) -> Table:
    table = Table("k", "v", title=f"file info: {file.name}", show_header=False)

    table.add_row("name", file.name)
    table.add_row("id", Text(str(file.id), style="green"))
    table.add_row("project", file.project_name)
    table.add_row("project id", Text(str(file.project_id), style="green"))
    table.add_row("mission", file.mission_name)
    table.add_row("mission id", Text(str(file.mission_id), style="green"))
    table.add_row("created", str(file.created_at))
    table.add_row("updated", str(file.updated_at))
    table.add_row("size", format_bytes(file.size))
    table.add_row("state", file_state_to_text(file.state))
    table.add_row("categories", ", ".join(file.categories))
    table.add_row("topics", ", ".join(file.topics))
    table.add_row("hash", file.hash)
    table.add_row("type", file.type_)
    table.add_row("date", str(file.date))

    return table


def mission_info_table(
    mission: Mission, print_metadata: bool = False
) -> Tuple[Table, ...]:
    table = Table("k", "v", title=f"mission info: {mission.name}", show_header=False)

    # TODO: add more fields as we store more information in the Mission object
    table.add_row("name", mission.name)
    table.add_row("id", Text(str(mission.id), style="green"))
    table.add_row("project", mission.project_name)
    table.add_row("project id", Text(str(mission.project_id), style="green"))
    table.add_row("created", str(mission.created_at))
    table.add_row("updated", str(mission.updated_at))
    table.add_row("size", format_bytes(mission.size))
    table.add_row("files", str(mission.number_of_files))

    if not print_metadata:
        return (table,)

    metadata_table = Table("k", "v", title="mission metadata", show_header=False)
    kv_pairs_sorted = sorted(
        [(k, v) for k, v in mission.metadata.items()], key=lambda x: x[0]
    )
    for k, v in kv_pairs_sorted:
        metadata_table.add_row(k, str(parse_metadata_value(v)))

    return table, metadata_table


def project_info_table(project: Project) -> Table:
    table = Table("k", "v", title=f"project info: {project.name}", show_header=False)

    # TODO: add more fields as we store more information in the Project object
    table.add_row("id", Text(str(project.id), style="green"))
    table.add_row("name", project.name)
    table.add_row("description", project.description)
    table.add_row("created", str(project.created_at))
    table.add_row("updated", str(project.updated_at))

    return table


def file_verification_status_table(
    file_status: Mapping[Path, FileVerificationStatus],
) -> Table:
    table = Table(title="file status")
    table.add_column("filename", style="cyan")
    table.add_column("status", style="green")
    for path, status in file_status.items():
        table.add_row(str(path), file_verification_status_to_text(status))
    return table


def print_file_verification_status(
    file_status: Mapping[Path, FileVerificationStatus], *, pprint: bool
) -> None:
    """\
    prints the file verification status to stdout / stderr
    either using pprint or as a list for piping
    """
    if pprint:
        table = file_verification_status_table(file_status)
        Console().print(table)
    else:
        for path, status in file_status.items():
            stream = (
                sys.stdout if status == FileVerificationStatus.UPLOADED else sys.stderr
            )
            print(path, file=stream, flush=True)


def print_files(files: Sequence[File], *, pprint: bool) -> None:
    """\
    prints the files to stdout / stderr
    either using pprint or as a list for piping
    """
    if pprint:
        table = files_to_table(files)
        Console().print(table)
    else:
        for file in files:
            stream = sys.stdout if file.state == FileState.OK else sys.stderr
            print(file.id, file=stream, flush=True)


def print_missions(missions: Sequence[Mission], *, pprint: bool) -> None:
    """\
    prints the missions to stdout
    either using pprint or as a list for piping
    """
    if pprint:
        table = missions_to_table(missions)
        Console().print(table)
    else:
        for mission in missions:
            print(mission.id)


def print_projects(projects: Sequence[Project], *, pprint: bool) -> None:
    """\
    prints the projects to stdout
    either using pprint or as a list for piping
    """
    if pprint:
        table = projects_to_table(projects)
        Console().print(table)
    else:
        for project in projects:
            print(project.id)


def print_file_info(file: File, *, pprint: bool) -> None:
    """\
    prints the file info to stdout
    either using pprint or as a list for piping
    """
    if pprint:
        table = file_info_table(file)
        Console().print(table)
    else:
        file_dct = asdict(file)
        for key in file_dct:
            file_dct[key] = str(file_dct[key])  # TODO: improve this
        print(json.dumps(file_dct))


def print_mission_info(mission: Mission, *, pprint: bool) -> None:
    """\
    prints the mission info to stdout
    either using pprint or as a list for piping
    """
    if pprint:
        Console().print(*mission_info_table(mission, print_metadata=True))
    else:
        mission_dct = asdict(mission)
        for key in mission_dct:
            mission_dct[key] = str(mission_dct[key])  # TODO: improve this
        print(json.dumps(mission_dct))


def print_project_info(project: Project, *, pprint: bool) -> None:
    """\
    prints the project info to stdout
    either using pprint or as a list for piping
    """
    if pprint:
        Console().print(project_info_table(project))
    else:
        project_dct = asdict(project)
        for key in project_dct:
            project_dct[key] = str(project_dct[key])  # TODO: improve this
        print(json.dumps(project_dct))



from __future__ import annotations

import time
from pathlib import Path
from secrets import token_hex

import pytest

from kleinkram import create_mission
from kleinkram import create_project
from kleinkram import delete_project
from kleinkram import list_missions
from kleinkram import list_projects
from kleinkram import upload

# we expect the mission files to be in this folder that is not commited to the repo
DATA_PATH = Path(__file__).parent.parent / "data" / "testing"
DATA_FILES = [
    DATA_PATH / "10_KB.bag",
    DATA_PATH / "50_KB.bag",
    DATA_PATH / "1_MB.bag",
    DATA_PATH / "17_MB.bag",
    DATA_PATH / "125_MB.bag",
]

PROJECT_DESCRIPTION = "This is a test project"


WAIT_BEOFORE_DELETION = 5


@pytest.fixture(scope="session")
def project():
    project_name = token_hex(8)

    create_project(project_name, description="This is a test project")
    project = list_projects(project_names=[project_name])[0]

    yield project

    time.sleep(WAIT_BEOFORE_DELETION)
    delete_project(project.id)


@pytest.fixture
def mission(project):
    mission_name = token_hex(8)
    upload(
        mission_name=mission_name,
        project_id=project.id,
        files=DATA_FILES,
        create=True,
    )
    mission = list_missions(project_ids=[project.id], mission_names=[mission_name])[0]

    yield mission


@pytest.fixture
def empty_mission(project):
    mission_name = token_hex(8)
    create_mission(mission_name, project.id, metadata={})
    mission = list_missions(project_ids=[project.id], mission_names=[mission_name])[0]

    yield mission
